#include "es.h"
#include "syscalls.h"
#include <stdio.h>		/* Pour PErreur */

static VTOC		vtoc;	/* Le systeme possede toujours une copie */
				/* de la VTOC */
static int		etat_vtoc;
#define VTOC_NON_CHARGEE	0
#define VTOC_CHARGEE		1

static Descripteur	descripteurs[MAX_DESCRIPTEURS]; /* Le systeme ne */
				/* peut pas ouvrir plus de MAX_DESCRIPTEURS */
				/* fichiers en meme temps!!! */

static Tampon_Courant	tampons[MAX_TAMPONS]; /* Le systeme */
				/* tampone MAX_TAMPONS secteurs */

void	PErreur(code_erreur erreur)
{
	static char	*messages[] = {
		"Pas d'erreur",		/* Ne devrait jamais apparaitre!!! */
		"Numero de descripteur impossible",
		"Mode d'ouverture inconnu",
		"Fichier inexistant",
		"Fichier deja ouvert",
		"Parametres faux",
		"Ecriture non autorisee",
		"Mauvaise position",
		"Catalog plein",
		"Fin de fichier",
		"Fichier existant",
		"Mauvais parametres",
		"Fichier ouvert",
		"Fichier protege",
		"Mauvais nom",
		"Fichier pas en ecriture",
		"Disque plein",
		"Fichier pas en lecture"
	};

	if (erreur & 0xff00) {
		erreur >>= 8;
		if (erreur<0 || erreur>MAX_MESSAGES)
			fprintf(stderr,"Erreur (DOS) Inconnue %x\n",erreur<<8);
		else
			fprintf(stderr,"%s\n",messages[erreur]);
	}
	else
		Perreur(erreur & 0x00ff);
}

static code_erreur	CloseFiles()
{
	int		i;
	Descripteur	*desc;
	code_erreur	erreur;

	/* On ferme les fichiers ouverts */
	fprintf(stderr,"Fermeture des fichiers ouverts: ");
	for (i=1,desc=descripteurs; i<=MAX_DESCRIPTEURS; i++,desc++) {
		if (DESC_OCC(*desc)) {
			erreur = SYS_Close(i);
			if (erreur!=OK) {
				PErreur(erreur);
				fprintf(stderr,"\nFermeture canal impossible");
				fprintf(stderr," %d (%s)\n",i,
					desc->localisation.fichier.nom);
			}
			else
				printf("%d ",i);
		}
	}
	printf("\n");
	return PAS_D_ERREUR;
}


static code_erreur	LectureVTOC()
{
	code_erreur	erreur;
	TS		adresse;

	if (etat_vtoc==VTOC_NON_CHARGEE) {	
		adresse.piste = PISTE_VTOC;
		adresse.secteur = SECTEUR_VTOC;
		erreur = LitSecteur(adresse,(octet *)&vtoc);
		if (erreur!=OK) return erreur;
		etat_vtoc = VTOC_CHARGEE;
	}
	return PAS_D_ERREUR;
}

static code_erreur	EcritureVTOC()
{
	TS	adresse;

	adresse.piste = PISTE_VTOC;
	adresse.secteur = SECTEUR_VTOC;
	return EcritSecteur(adresse,(octet *)&vtoc);
}

static code_erreur	LectureTampon(TS adresse,int *tc)
{
	Tampon_Courant	*tampon;
	code_erreur	erreur;

	/* Hashe... */
	tampon = tampons + (*tc = HASH(adresse));

	/* La place est prise */
	if (tampon->etat!=TAMPON_LIBRE) {
		/* Mais par le meme */
		if (MEME_TS(tampon->adresse,adresse)) return PAS_D_ERREUR;
		/* Ou pas par le meme et en plus "sale" alors on l'ecrit */
		if (tampon->etat==TAMPON_SALE) {
			erreur = EcritSecteur(tampon->adresse,
				(octet *)&tampon->tampon);
			if (erreur!=OK) return erreur;
		}
	}
	/* Cette fois on peut charger */
	tampon->etat = TAMPON_PROPRE;
	tampon->adresse = adresse;
	return LitSecteur(adresse,(octet *)&tampon->tampon);
}

static code_erreur	EcritureTampon(int tampon)
{
	tampons[tampon].etat = TAMPON_SALE;
	return PAS_D_ERREUR;
}

static code_erreur	ChercheFichierOuvert(Localisation *loc)
{
	code_erreur	erreur;
	int		i;
	Descripteur	*desc;

	/* Pour l'instant c'est libre */
	loc->occupation = LOCALISATION_LIBRE;

	/* On cherche les places occupees avec le meme nom */
	for (i=0,desc=descripteurs; i<MAX_DESCRIPTEURS; i++,desc++)
		if (DESC_OCC(*desc) &&
		    !strcmp(desc->localisation.fichier.nom,loc->fichier.nom)) {
			*loc = desc->localisation;
			break;
		}
	return PAS_D_ERREUR;
}

static code_erreur	ChercheFichierDisque(Localisation *loc)
{
	TS		adresse;
	code_erreur	erreur;
	Fichier		*fichier;
	int		cc;		/* Catalog Courant */
	int		nfic;

	loc->occupation = PLUS_DE_LOCALISATION;
	adresse = vtoc.catalog;
	/* Tant qu'on n'est pas en fin de chainage de catalog */
	while (!FIN_CHAINAGE(adresse)) {
		/* On charge un catalogue */
		erreur = LectureTampon(adresse,&cc);
		if (erreur!=OK) return erreur;

		/* On regarde tout les fichiers */
		for (nfic=0,fichier=tampons[cc].tampon.catalog.fichier;
			nfic<FICHIERS_PAR_CATALOG; nfic++,fichier++) {
			if (FICHIER_LIBRE(*fichier)) {
				/* On se souvient qu'il y avait une place */
				/* libre ici ca pourrait servir */
				if (loc->occupation==PLUS_DE_LOCALISATION) {
					loc->occupation = LOCALISATION_LIBRE;
					loc->catalog = adresse;
					loc->entree = nfic;
					loc->fichier = *fichier;
				}
			}
			else
				/* On l'a trouve */
				if (!strncmp(loc->fichier.nom,fichier->nom,
					LONGUEUR_NOM)) {
					loc->occupation = LOCALISATION_OCCUPEE;
					loc->catalog = adresse;
					loc->entree = nfic;
					loc->fichier = *fichier;
					return PAS_D_ERREUR;
				}
		}
		/* On continue le chainage */
		adresse = tampons[cc].tampon.catalog.suivant;
	}
	return PAS_D_ERREUR;
}

code_erreur	SYS_Init(octet volume)
{
	Catalog			catalog;
	TS			adresse;
	register int		nfic;
	code_erreur		erreur;
	register int		i;

	/* Il faut marquer les tampons comme etant inutilises... */
	for (i=0; i<MAX_TAMPONS; i++)
		tampons[i].etat = TAMPON_LIBRE;

	/* Et marquer les descripteurs comme fermes... */
	for (i=0; i<MAX_DESCRIPTEURS; i++)
		descripteurs[i].localisation.occupation = LOCALISATION_LIBRE;

	/* Ensuite on s'occupe de fabriquer la VTOC */
	bzero(&vtoc,sizeof(VTOC));
	vtoc.catalog.piste = PISTE_CATALOG;
	vtoc.catalog.secteur = DEBUT_CATALOG_S;
	vtoc.version = VERSION_DOS;
	vtoc.volume = volume;
	vtoc.n_ts = TS_PAR_TSL;
	vtoc.np = N_PISTES;
	vtoc.ns = N_SECTEURS;
	vtoc.no = N_OCTETS;
	vtoc.bitmap[0x00][0] = PISTE_OCCUPEE;	/* Boot+RWTS+Re+Kernel */
	vtoc.bitmap[0x01][0] = PISTE_OCCUPEE;	/* Kernel */
	vtoc.bitmap[0x02][0] = 0xfff0;		/* Kernel puis libre */
	for (adresse.piste=0x03; adresse.piste<PISTE_CATALOG; adresse.piste++)
		vtoc.bitmap[adresse.piste][0] = PISTE_LIBRE;
	vtoc.bitmap[PISTE_CATALOG][0] = PISTE_OCCUPEE;
	for (adresse.piste=PISTE_CATALOG+1; adresse.piste<N_PISTES;
	     adresse.piste++)
		vtoc.bitmap[adresse.piste][0] = PISTE_LIBRE;
	erreur = EcritureVTOC();
	if (erreur!=OK) return erreur;
	/* Ensuite on fabrique les secteurs de CATALOG chaines entre eux */
	bzero(&catalog,sizeof(Catalog));
	catalog.suivant.piste = PISTE_CATALOG;
	for (nfic=0; nfic<FICHIERS_PAR_CATALOG; nfic++)
		LIBERE_FICHIER(catalog.fichier[nfic]);
	adresse.piste = PISTE_CATALOG;
	for (adresse.secteur=DEBUT_CATALOG_S; adresse.secteur>FIN_CATALOG_S;
	     adresse.secteur--) {
		catalog.suivant.secteur = adresse.secteur-1;
		erreur = EcritSecteur(adresse,(octet *)&catalog);
		if (erreur!=OK) return erreur;
	}
	catalog.suivant.piste = 0;
	catalog.suivant.secteur = 0;
	adresse.piste = PISTE_CATALOG;
	adresse.secteur = FIN_CATALOG_S;
	erreur = EcritSecteur(adresse,(octet *)&catalog);
	if (erreur!=OK) return erreur;
	return PAS_D_ERREUR;
}

code_erreur	SYS_Position(int canal,int position,int mode)
{
	code_erreur	erreur;
	int		npos, its, vits, nt, ntsl;
	TS		adresse;
	Descripteur	*desc;

	/* Verification de la validite des parametres */
	if (!BON_CANAL(canal)) return MAUVAIS_DESCRIPTEUR;
	canal--;
	desc = descripteurs+canal;
	if (!DESC_OCC(*desc)) return MAUVAIS_DESCRIPTEUR;
	switch (mode) {
		default:
		case POSITION_DEBUT:
			npos = position;
			break;
		case POSITION_FIN:
			npos = desc->localisation.fichier.taille-position;
			break;
		case POSITION_COURANTE:
			npos = desc->deplacement+position;
			break;
	}
	if (npos<0 || npos>desc->localisation.fichier.taille)
		return MAUVAISE_POSITION;
	its = npos/TAILLE_SECTEUR;
	vits = desc->deplacement/TAILLE_SECTEUR;
	/* On est deja dans la meme page!!! */
	if (its==vits) {
		desc->deplacement = npos;
		return PAS_D_ERREUR;
	}
	/* Est-ce qu'on est dans la meme TSL ? */
	if (its/TS_PAR_TSL==vits/TS_PAR_TSL) {
		/* Charge la TSL */
		erreur = LectureTampon(desc->tsl,&nt);
		if (erreur!=OK) return erreur;
		/* Pointe sur le secteur */
		desc->secteur = tampons[nt].tampon.tsl.ts[its%TS_PAR_TSL];
		desc->ts_courante = its;
		desc->deplacement = npos;
		return PAS_D_ERREUR;
	}
	/* Il faut revenir en arriere ou en avant ? */
	if (its<vits) {	/* En arriere */
		ntsl = its/TS_PAR_TSL;
		erreur = LectureTampon(desc->localisation.catalog,&nt);
		if (erreur!=OK) return erreur;
		adresse = tampons[nt].tampon.catalog.fichier[
			desc->localisation.entree].tsl;
	}
	else {
		ntsl = (vits/TS_PAR_TSL)-(its/TS_PAR_TSL)-1;
		adresse = desc->tsl;
	}
	while (ntsl--) {
		erreur = LectureTampon(adresse,&nt);
		if (erreur!=OK) return erreur;
		adresse = tampons[nt].tampon.tsl.suivant;
	}
	desc->tsl = adresse;
	erreur = LectureTampon(desc->tsl,&nt);
	if (erreur!=OK) return erreur;
	desc->ts_courante = its;
	desc->secteur = tampons[nt].tampon.tsl.ts[its];
	desc->deplacement = npos;
	return PAS_D_ERREUR;
}

code_erreur	SYS_Open(char *nom,int mode,int canal)
{
	Descripteur	*desc;
	code_erreur	erreur;
	int		nt;
	Localisation	loc;

	/* Verification de la validite des parametres */
	if (!BON_CANAL(canal)) return MAUVAIS_DESCRIPTEUR;
	canal--;			/* Vive les tableaux en C */
	if (nom==NULL) return MAUVAIS_NOM;
	if (mode<0 || mode>3) return MODE_INCONNU;

	/* Est-ce que le canal est utilise ? */
	desc = descripteurs+canal;
	if (DESC_OCC(*desc)) return DEJA_OUVERT;

	/* Est-ce que le fichier existe ? */
	strncpy(loc.fichier.nom,nom,LONGUEUR_NOM);
	erreur = ChercheFichierOuvert(&loc);
	if (erreur!=OK) return erreur;
	if (loc.occupation==LOCALISATION_OCCUPEE) return DEJA_OUVERT;

	erreur = ChercheFichierDisque(&loc);
	if (erreur!=OK) return erreur;
	if (loc.occupation!=LOCALISATION_OCCUPEE) return FICHIER_INEXISTANT;

	/* Bon on peut tenter de l'ouvrir maintenant */
	desc->mode = mode;
	desc->localisation = loc;
	desc->tsl = desc->localisation.fichier.tsl;
	if (!PAS_DE_TSL(*desc)) {
		erreur = LectureTampon(desc->tsl,&nt);
		if (erreur!=OK) return erreur;
		desc->secteur = tampons[nt].tampon.tsl.ts[0];
	}
	else {
		desc->secteur.piste = 0;
		desc->secteur.secteur = 0;
	}
	desc->deplacement = 0;
	desc->nbre_ts = desc->localisation.fichier.taille/TAILLE_SECTEUR +
		(desc->localisation.fichier.taille%TAILLE_SECTEUR)?1:0;
	desc->ts_courante = 0;
	return PAS_D_ERREUR;
}

code_erreur	ChercheBloc(TS *bloc)
{
	code_erreur	erreur;
	TS		adresse;

	/* Regarde un peu la bitmap */
	for (adresse.piste=0; adresse.piste<N_PISTES; adresse.piste++) {
		if (vtoc.bitmap[adresse.piste][0]==PISTE_OCCUPEE) continue;
		for (adresse.secteur=0; adresse.secteur<N_SECTEURS;
		     adresse.secteur++) {
			if (BLOC_LIBRE(vtoc.bitmap,adresse)) {
				*bloc = adresse;
				PRENDS_BLOC(vtoc.bitmap,adresse);
				erreur = EcritureVTOC();
				if (erreur!=OK) return erreur;
				return PAS_D_ERREUR;
			}
		}
	}
	return DISQUE_PLEIN;
}

static code_erreur	AjouteTSL(Descripteur *desc)
{
	code_erreur	erreur;
	TS		tsl;
	int		nt, nf;

	/* Recupere un bloc */
	erreur = ChercheBloc(&tsl);
	if (erreur!=OK) return erreur;

	/* Marque le avec une fin de chainage */
	erreur = LectureTampon(tsl,&nt);
	if (erreur!=OK) return erreur;
	tampons[nt].tampon.tsl.suivant.piste = 0;
	tampons[nt].tampon.tsl.suivant.secteur = 0;
	erreur = EcritureTampon(nt);
	if (erreur!=OK) return erreur;

	if (PAS_DE_TSL(*desc)) {	/* C'est la premiere */
		erreur = LectureTampon(desc->localisation.catalog,&nt);
		if (erreur!=OK) return erreur;
		nf = desc->localisation.entree;
		tampons[nt].tampon.catalog.fichier[nf].tsl = tsl;
		desc->localisation.fichier.tsl = tsl;
	}
	else {	/* Fait un chainage avec la precedente */
		erreur = LectureTampon(desc->tsl,&nt);
		if (erreur!=OK) return erreur;
		tampons[nt].tampon.tsl.suivant = tsl;
	}
	erreur = EcritureTampon(nt);
	if (erreur!=OK) return erreur;

	/* Maintenant on pointe sur le secteur de tsl */
	desc->tsl = tsl;
	return PAS_D_ERREUR;
}

static code_erreur	EcritFichier(Descripteur *desc,octet donnee)
{
	code_erreur	erreur;
	TS		secteur;
	int		nts, ntt;
	TSL		tsl;
	int		tsc;

	/* IL FAUDRAIT RECUPERER LES ERREURS MAIS... */

	/* Je rajoute ou j'ecrase ? */
	if (desc->deplacement==desc->localisation.fichier.taille) {
		if (desc->localisation.fichier.taille%TAILLE_SECTEUR==0) {
			if (PAS_DE_TSL(*desc) ||
			    desc->nbre_ts%TS_PAR_TSL==0) {	/* Ajout TSL */
				erreur = AjouteTSL(desc);
				if (erreur!=OK) return erreur;
			}
			/* Recupere un bloc */
			erreur = ChercheBloc(&secteur);
			if (erreur!=OK) return erreur;

			/* On rajoute une TS */
			tsc = desc->ts_courante%TS_PAR_TSL;
			erreur = LectureTampon(desc->tsl,&ntt);
			if (erreur!=OK) return erreur;
			tampons[ntt].tampon.tsl.ts[tsc] = secteur;
			erreur = EcritureTampon(ntt);
			if (erreur!=OK) return erreur;
			desc->nbre_ts++;
			desc->ts_courante++;

			/* Maintenant on pointe sur le secteur de donnees */
			desc->secteur = secteur;
		}
		desc->localisation.fichier.taille++;
	}
	/* On charge le secteur de donnees */
	secteur = desc->secteur;
	erreur = LectureTampon(secteur,&nts);
	if (erreur!=OK) return erreur;

	/* On peut rajouter l'octet a l'offset indique */
	tampons[nts].tampon.secteur[desc->deplacement%TAILLE_SECTEUR] = donnee;
	erreur = EcritureTampon(nts);
	if (erreur!=OK) return erreur;

	/* Maintenant on mets a jour l'offset */
	desc->deplacement++;

	return PAS_D_ERREUR;
}

static code_erreur	LitFichier(Descripteur *desc,octet *donnee)
{
	code_erreur	erreur;
	int		nt;
	TS		secteur;

	/* Est-ce qu'on est en fin de fichier ? */
	if (desc->localisation.fichier.taille==desc->deplacement)
		return FIN_DE_FICHIER;

	/* On charge le secteur de donnees */
	secteur = desc->secteur;
	erreur = LectureTampon(secteur,&nt);
	if (erreur!=OK) return erreur;
	*donnee = tampons[nt].tampon.secteur[desc->deplacement%TAILLE_SECTEUR];
	/* On en a lu 1 */
	desc->deplacement++;

	/* Faut-il changer de TS ? */
	if (desc->deplacement%TAILLE_SECTEUR==0) {
		desc->ts_courante++;
		/* Faut-il changer de TSL ? */
		if (desc->ts_courante%TS_PAR_TSL==0) {
			erreur = LectureTampon(desc->tsl,&nt);
			if (erreur!=OK) return erreur;
			desc->tsl = tampons[nt].tampon.tsl.suivant;
		}
		/* On charge la TSL */
		erreur = LectureTampon(desc->tsl,&nt);
		if (erreur!=OK) return erreur;
		/* On lit la TS */
		desc->secteur = tampons[nt].tampon.tsl.ts[
			desc->ts_courante%TS_PAR_TSL];
	}
	return PAS_D_ERREUR;
}

code_erreur	SYS_Read(int canal,int *longueur,octet *donnees)
{
	code_erreur	erreur;
	Descripteur	*desc;
	int		i;

	/* Verification de la validite des parametres */
	if (!BON_CANAL(canal)) return MAUVAIS_DESCRIPTEUR;
	canal--;
	desc = descripteurs+canal;
	if (desc->localisation.occupation!=DESCRIPTEUR_OCCUPE)
		return MAUVAIS_DESCRIPTEUR;
	if (!(desc->mode&LECTURE)) return FICHIER_PAS_EN_LECTURE;

	for (i=0; i<*longueur; i++) {
		erreur = LitFichier(desc,donnees+i);
		if (erreur!=OK) {
			*longueur = i;
			return erreur;
		}
	}
	return PAS_D_ERREUR;
}

code_erreur	SYS_Write(int canal,octet *tampon,mot taille)
{
	code_erreur	erreur, err;
	Descripteur	*desc;
	int		i, nt;
	mot		otaille;

	/* Verification de la validite des parametres */
	if (!BON_CANAL(canal)) return MAUVAIS_DESCRIPTEUR;
	canal--;
	desc = descripteurs+canal;
	if (desc->localisation.occupation!=DESCRIPTEUR_OCCUPE)
		return MAUVAIS_DESCRIPTEUR;
	if (!(desc->mode&ECRITURE)) return FICHIER_PAS_EN_ECRITURE;

	/* On se souvient de l'ancienne taille */
	otaille = desc->localisation.fichier.taille;

	/* Bon ben il faut essayer d'ecrire maintenant!!! */
	for (i=0; i<taille; i++) {
		err = EcritFichier(desc,*tampon++);
		if (err!=OK) break;
	}
	if (otaille<desc->localisation.fichier.taille) {
		erreur = LectureTampon(desc->localisation.catalog,&nt);
		if (erreur!=OK) return erreur;
		tampons[nt].tampon.catalog.fichier[
			desc->localisation.entree].taille =
			desc->localisation.fichier.taille;
		erreur = EcritureTampon(nt);
		if (erreur!=OK) return erreur;
		if (err!=OK) return err;
		return PAS_D_ERREUR;
	}
	if (err!=OK) return err;
	return PAS_D_ERREUR;
}

code_erreur	SYS_Close(int canal)
{
	Descripteur	*desc;

	/* Verification de la validite des parametres */
	if (!BON_CANAL(canal)) return MAUVAIS_DESCRIPTEUR;
	canal--;
	desc = descripteurs+canal;
	if (!DESC_OCC(*desc)) return MAUVAIS_DESCRIPTEUR;
	desc->localisation.occupation = LOCALISATION_LIBRE;
	return PAS_D_ERREUR;
}

code_erreur	SYS_Catalog()
{
	TS		adresse;
	code_erreur	erreur;
	int		cc;
	register int	nfic;
	struct Fichier	*fichier;
	int		compte;

	printf("\n\n\nCatalog DOS 3.%d Volume numero %d\n",vtoc.version,
		vtoc.volume);
	compte = 0;
	adresse = vtoc.catalog;
	while (!FIN_CHAINAGE(adresse)) {
		erreur = LectureTampon(adresse,&cc);
		if (erreur!=OK) return erreur;
		for (nfic=0,fichier=tampons[cc].tampon.catalog.fichier;
		     nfic<FICHIERS_PAR_CATALOG; nfic++,fichier++) {
			if (!FICHIER_LIBRE(*fichier)) {
				printf("%6d %c %c %s\n",
					fichier->taille,
					EST_VERROUILLE(*fichier)?'V':' ',
					"TBAxbxxxSxxxxxxxR"[fichier->type&0x3f],
					fichier->nom);
				compte++;
			}
		adresse = tampons[cc].tampon.catalog.suivant;
		}
	}
	if (compte==0)
		printf("Le catalog ne contient pas de fichier\n\n\n\n");
	else
		printf("Le catalog contient %d fichier%c\n\n\n\n",
			compte,compte>1?'s':' ');
	return PAS_D_ERREUR;
}

code_erreur	SYS_InitialisationSysteme()
{
	code_erreur	erreur;
	register int	i;

	erreur = DemarreMoteur();
	if (erreur!=OK) {
		PErreur(erreur);
		return erreur;
	}

	/* Tout les tampons sont libres */
	for (i=0; i<MAX_TAMPONS; i++)
		tampons[i].etat = TAMPON_LIBRE;

	/* Tout les descripteurs sont fermes */
	for (i=0; i<MAX_DESCRIPTEURS; i++)
		descripteurs[i].localisation.occupation = LOCALISATION_LIBRE;

	/* Charge la VTOC */
	erreur = LectureVTOC();
	if (erreur!=OK) return erreur;

	fprintf(stderr,"Emulation APPLE DOS 3.3 File System (c)1994 JBY\n");
	fprintf(stderr,"-----------------------------------------------\n");
	fprintf(stderr,"Lancement du systeme\n");
	fprintf(stderr,"Nombre de secteurs tampon: %d\n",MAX_TAMPONS);
	fprintf(stderr,"Nombre de canaux disponibles: %d\n",MAX_DESCRIPTEURS);
	fprintf(stderr,"Taille des donnees statiques du noyau: %d+%d+%d",
		sizeof(vtoc),sizeof(tampons),sizeof(descripteurs));
	fprintf(stderr," octets\n\n\n");
	return OK;
}

code_erreur	SYS_TerminaisonSysteme()
{
	code_erreur	erreur;
	register int	i;
	Descripteur	*desc;

	fprintf(stderr,"\n\n\nExtinction du systeme\n");

	/* On ferme les fichiers ouverts */
	erreur = CloseFiles();
	if (erreur!=OK) PErreur(erreur);;

	/* On vide les secteurs tampons */
	fprintf(stderr,"Vidage des secteurs tampons du systeme: ");
	for (i=0; i<MAX_TAMPONS; i++)
		if (tampons[i].etat == TAMPON_SALE) {
			erreur = EcritSecteur(tampons[i].adresse,
				(octet *)&tampons[i].tampon);
			if (erreur!=OK) {
				PErreur(erreur);
				fprintf(stderr,"\nVidage buffer impossible");
				fprintf(stderr,"%d: <%2x,%2x>\n",i,
					tampons[i].adresse.piste,
					tampons[i].adresse.secteur);
			}
			else
				printf("(%d:<%2x,%2x>) ",i,
					tampons[i].adresse.piste,
					tampons[i].adresse.secteur);
		}
	printf("\n");


	/* On peut arreter le disque */
	erreur = EteintMoteur();
	if (erreur!=OK) {
		PErreur(erreur);
		return erreur;
	}
	return OK;
}

code_erreur	SYS_ShowOpen()
{
	int		i;
	Descripteur	*desc;

	printf("\n\n\nListe des fichiers ouverts\n");
	printf("De Mo Taille(Deplac) Pi Se En TSLco TSLin Secte NTS TSC Nom\n");
	for (i=0,desc=descripteurs; i<MAX_DESCRIPTEURS; i++,desc++) {
		if (DESC_OCC(*desc)) {
			printf("%2d %c%c %6d(%6d) %2x %2x %2x %2x %2x %2x %2x",
				i+1,
				desc->mode&LECTURE?'L':' ',
				desc->mode&ECRITURE?'E':' ',
				desc->localisation.fichier.taille,
				desc->deplacement,
				desc->localisation.catalog.piste,
				desc->localisation.catalog.secteur,
				desc->localisation.entree,
				desc->tsl.piste,
				desc->tsl.secteur,
				desc->localisation.fichier.tsl.piste,
				desc->localisation.fichier.tsl.secteur);
			printf(" %2x %2x %3d %3d %s\n",
				desc->secteur.piste,
				desc->secteur.secteur,
				desc->nbre_ts,
				desc->ts_courante,
				desc->localisation.fichier.nom);
		}
	}
	printf("\n\n\n");
	return PAS_D_ERREUR;
}

code_erreur	SYS_ShowSector(num_piste np,num_secteur ns,char *t)
{
	code_erreur	erreur;
	int		is, i, j;
	char		*sect;
	Secteur		*secteur;
	Catalog		*catalog;
	Fichier		*fichier;
	VTOC		*vtoc;
	TS		adresse;

	/* Verification des parametres */
	if (np>MAX_PISTES || ns>MAX_SECTEURS) return MAUVAIS_PARAMETRES;

	/* On essaie de lire le secteur correspondant */
	adresse.piste = np;
	adresse.secteur = ns;
	erreur = LectureTampon(adresse,&is);
	if (erreur!=OK) return erreur;
	switch(*t) {
		case 'N':
			secteur = &tampons[is].tampon.secteur;
			printf("\n\n\n   :");
			for (j=0; j<16; j++) {
				printf(" %02x",j);
			}
			printf("\n");
			for (i=0; i<16; i++) {
				printf("%02x :",i*16);
				for (j=0; j<16; j++) {
					printf(" %02x",(*secteur)[i*16+j]);
				}
				printf(" :",i*16);
				for (j=0; j<16; j++) {
					printf("%c",isalpha((*secteur)[i*16+j])?
						(*secteur)[i*16+j]:'.');
				}
				printf("\n");
			}
			printf("\n\n\n");
			break;
		case 'V':
			vtoc = (VTOC *)&tampons[is].tampon.secteur;
			printf("\n\n\n");
			printf("Premier catalogue en piste %x et secteur %x\n",
				vtoc->catalog.piste,vtoc->catalog.secteur);
			printf("Version du DOS %d\n",vtoc->version);
			printf("Nombre de TS pas TSL: %d\n",vtoc->n_ts);
			printf("Nombre de pistes par disque: %d\n",vtoc->np);
			printf("Nombre de secteurs par piste: %d\n",vtoc->ns);
			printf("Nombre d'octets par secteur: %d\n",vtoc->no);
			printf("Disque de capacite: %d\n",
				vtoc->np*vtoc->ns*vtoc->no);
			printf("Bitmap:\n  |");
			for (i=0; i<N_PISTES; i++) printf("%1x|",i/16);
			printf("\n  |");
			for (i=0; i<N_PISTES; i++) printf("%1x|",i%16);
			printf("\n--|");
			for (i=0; i<N_PISTES; i++) printf("-+");
			printf("\n");
			for (j=0; j<N_SECTEURS; j++) {
				printf("%02x|",j);
				adresse.secteur = j;
				for (i=0; i<N_PISTES; i++) {
					adresse.piste = i;
					printf("%c|",BLOC_LIBRE(vtoc->bitmap,
						adresse)?' ':'X');
				}
				printf("\n");
			}
			printf("--+");
			for (i=0; i<N_PISTES; i++) printf("-+");
			printf("\n");
			printf("\n\n\n");
			break;
		case 'C':
			catalog = &tampons[is].tampon.catalog;
			printf("\n\n\n");
			printf("Catalogue suivant en piste %x et secteur %x\n",
			       catalog->suivant.piste,catalog->suivant.secteur);
			for (i=0,fichier=catalog->fichier;
			     i<FICHIERS_PAR_CATALOG; i++,fichier++) {
				printf("Fichier %2d: ",i);
				if (!FICHIER_LIBRE(catalog->fichier[i])) {
					printf("TSL: %2x %2x Type: %d ",
						fichier->tsl.piste,
						fichier->tsl.secteur,
						fichier->type);
					printf("Taille: %5d Nom: %s\n",
						fichier->taille,
						fichier->nom);
				}
				else {
					if (!FIN_CHAINAGE(fichier->tsl) ||
					    fichier->type!=0 ||
					    fichier->taille!=0 ||
					    fichier->nom[1]!='\0') {
						printf("TSL: %2x %2x Type: %d ",
							fichier->tsl.piste,
							fichier->tsl.secteur,
							fichier->type);
						printf("Taille: %5d Nom: ?%s ",
							fichier->taille,
							fichier->nom+1);
						printf(" (Entree effacee)\n");
					}
					else
						printf("Entree libre\n");
				}
			}
			printf("\n\n\n");
			break;
	}
	return PAS_D_ERREUR;
}

code_erreur	SYS_Rename(char *ancien,char *nouveau)
{
	Localisation	loc_ancien, loc_nouveau;
	int		nt;
	code_erreur	erreur;

	/* Verification des parametres */
	if (ancien==NULL || nouveau==NULL) return MAUVAIS_NOM;

	/* Est-ce que le premier fichier est ouvert ? */
	strncpy(loc_ancien.fichier.nom,ancien,LONGUEUR_NOM);
	erreur = ChercheFichierOuvert(&loc_ancien);
	if (erreur!=OK) return erreur;
	if (loc_ancien.occupation==LOCALISATION_OCCUPEE)
		return FICHIER_OUVERT;

	/* Est-ce que le premier fichier existe ? */
	erreur = ChercheFichierDisque(&loc_ancien);
	if (erreur!=OK) return erreur;
	if (loc_ancien.occupation!=LOCALISATION_OCCUPEE)
		return FICHIER_INEXISTANT;

	/* Est-ce que le deuxieme est ouvert ? */
	strncpy(loc_nouveau.fichier.nom,nouveau,LONGUEUR_NOM);
	erreur = ChercheFichierOuvert(&loc_nouveau);
	if (erreur!=OK) return erreur;
	if (loc_nouveau.occupation==LOCALISATION_OCCUPEE)
		return FICHIER_OUVERT;

	/* Est-ce que le deuxieme existe ? */
	erreur = ChercheFichierDisque(&loc_nouveau);
	if (erreur!=OK) return erreur;
	if (loc_nouveau.occupation==LOCALISATION_OCCUPEE)
		return FICHIER_EXISTANT;

	/* Bon maintenant on peut renommer */
	erreur = LectureTampon(loc_ancien.catalog,&nt);
	if (erreur!=OK) return erreur;
	strncpy(tampons[nt].tampon.catalog.fichier[loc_ancien.entree].nom,
		nouveau,LONGUEUR_NOM);
	erreur = EcritureTampon(nt);
	if (erreur!=OK) return erreur;
	return PAS_D_ERREUR;
}

code_erreur	SYS_Lock(char *nom)
{
	code_erreur	erreur;
	Localisation	loc;
	int		nt;

	/* Verification des parametres */
	if (nom==NULL) return MAUVAIS_NOM;

	/* Est-ce que le fichier est ouvert ? */
	strncpy(loc.fichier.nom,nom,LONGUEUR_NOM);
	erreur = ChercheFichierOuvert(&loc);
	if (erreur!=OK) return erreur;
	if (loc.occupation==LOCALISATION_OCCUPEE) return FICHIER_OUVERT;

	/* Est-ce que le fichier existe ? */
	erreur = ChercheFichierDisque(&loc);
	if (erreur!=OK) return erreur;
	if (loc.occupation!=LOCALISATION_OCCUPEE) return FICHIER_INEXISTANT;

	/* On peut modifier le bit de verouillage */
	erreur = LectureTampon(loc.catalog,&nt);
	if (erreur!=OK) return erreur;
	if (EST_VERROUILLE(tampons[nt].tampon.catalog.fichier[loc.entree]))
		return PAS_D_ERREUR;
	VERROUILLE(tampons[nt].tampon.catalog.fichier[loc.entree]);
	erreur = EcritureTampon(nt);
	if (erreur!=OK) return erreur;
	return PAS_D_ERREUR;
}

code_erreur	SYS_Unlock(char *nom)
{
	code_erreur	erreur;
	Localisation	loc;
	int		nt;

	/* Verification des parametres */
	if (nom==NULL) return MAUVAIS_NOM;

	/* Est-ce que le fichier est ouvert ? */
	strncpy(loc.fichier.nom,nom,LONGUEUR_NOM);
	erreur = ChercheFichierOuvert(&loc);
	if (erreur!=OK) return erreur;
	if (loc.occupation==LOCALISATION_OCCUPEE) return FICHIER_OUVERT;

	/* Est-ce que le fichier existe ? */
	erreur = ChercheFichierDisque(&loc);
	if (erreur!=OK) return erreur;
	if (loc.occupation!=LOCALISATION_OCCUPEE) return FICHIER_INEXISTANT;

	/* On peut modifier le bit de verouillage */
	erreur = LectureTampon(loc.catalog,&nt);
	if (erreur!=OK) return erreur;
	if (!EST_VERROUILLE(tampons[nt].tampon.catalog.fichier[loc.entree]))
		return PAS_D_ERREUR;
	DEVERROUILLE(tampons[nt].tampon.catalog.fichier[loc.entree]);
	erreur = EcritureTampon(nt);
	if (erreur!=OK) return erreur;
	return PAS_D_ERREUR;
}

code_erreur	SYS_Delete(char *nom)
{
	code_erreur	erreur;
	Localisation	loc;
	int		nt;
	TS		tsl;
	int		nbre_ts;
	Fichier		*fichier;
	TSL		*ptsl;
	mot		taille;
	int		i;
	TS		adresse;

	/* Verification des parametres */
	if (nom==NULL) return MAUVAIS_NOM;

	/* Le fichier est ouvert ? */
	strncpy(loc.fichier.nom,nom,LONGUEUR_NOM);
	erreur = ChercheFichierOuvert(&loc);
	if (erreur!=OK) return erreur;
	if (loc.occupation==LOCALISATION_OCCUPEE) return FICHIER_OUVERT;

	/* Le fichier existe ? */
	erreur = ChercheFichierDisque(&loc);
	if (erreur!=OK) return erreur;
	if (loc.occupation!=LOCALISATION_OCCUPEE) return FICHIER_INEXISTANT;

	/* Alors essaie de le detruire! */
	erreur = LectureTampon(loc.catalog,&nt);
	if (erreur!=OK) return erreur;
	fichier = tampons[nt].tampon.catalog.fichier+loc.entree;
	if (EST_VERROUILLE(*fichier)) return FICHIER_PROTEGE;
	LIBERE_FICHIER(*fichier);
	erreur = EcritureTampon(nt);
	if (erreur!=OK) return erreur;

	/* Ensuite on parcourt la liste des TSL pour nettoyer le tout!!! */
	tsl = fichier->tsl;
	/* Calcule le nombre de paires TS pour ce fichier */
	taille = fichier->taille;
	nbre_ts = taille/TAILLE_SECTEUR + (taille%TAILLE_SECTEUR)?1:0;
	while (!FIN_CHAINAGE(tsl)) {
		erreur = LectureTampon(tsl,&nt);
		if (erreur!=OK) return PAS_D_ERREUR;
		ptsl = &tampons[nt].tampon.tsl;
		for (i=0; i<TS_PAR_TSL; i++,nbre_ts--) {
			LIBERE_BLOC(vtoc.bitmap,ptsl->ts[i]);
			/* On vient de liberer le dernier */
			if (nbre_ts==1) {
				LIBERE_BLOC(vtoc.bitmap,tsl);
				erreur = EcritureVTOC();
				if (erreur!=OK) return erreur;
				return PAS_D_ERREUR;
			}
		}
		tsl = ptsl->suivant;
	}
	return PAS_D_ERREUR;
}

code_erreur	SYS_Create(char *nom)
{
	code_erreur	erreur;
	Localisation	loc;
	int		nt;
	Fichier		*fichier;

	/* Verification des parametres */
	if (nom==NULL) return MAUVAIS_NOM;

	/* Est ce qu'il est ouvert ? */
	strncpy(loc.fichier.nom,nom,LONGUEUR_NOM);
	erreur = ChercheFichierOuvert(&loc);
	if (erreur!=OK) return erreur;
	if (loc.occupation==LOCALISATION_OCCUPEE) return FICHIER_OUVERT;

	/*  Est-ce qu'il existe ? */
	erreur = ChercheFichierDisque(&loc);
	if (erreur!=OK) return erreur;
	switch (loc.occupation) {
		case PLUS_DE_LOCALISATION:
			return CATALOG_PLEIN;
		case LOCALISATION_OCCUPEE:
			return FICHIER_EXISTANT;
	}
	/* Il y a une place libre!!! */
	erreur = LectureTampon(loc.catalog,&nt);
	if (erreur!=OK) return erreur;
	fichier = tampons[nt].tampon.catalog.fichier+loc.entree;
	fichier->tsl.piste = 0;
	fichier->tsl.secteur = 0;
	fichier->type = FICHIER_TEXTE;
	strncpy(fichier->nom,nom,LONGUEUR_NOM);
	fichier->taille = 0;
	erreur = EcritureTampon(nt);
	if (erreur!=OK) return erreur;
	return PAS_D_ERREUR;
}
